[Version = "1.3.0"]
section DevResults.PowerBI;

// silly that we need to repeat the version here, but whatever man. It's handled by GitHub Actions
version = "1.3.0";
client_id = "Power BI Connector";

[DataSource.Kind = "DevResults.PowerBI", Publish = "DevResults.PowerBI.Publish"]
shared DevResults.PowerBI.Contents = Value.ReplaceType(NavTable, NavTableType);

NavTableType = type function (
    subdomain as (
        type text meta [
            // Parameter field label
            Documentation.FieldCaption = Extension.LoadString("SubdomainCaption"),
            // Hover info
            Documentation.FieldDescription = Extension.LoadString("SubdomainDescription"),
            // Watermark example
            Documentation.SampleValues = {Extension.LoadString("SubdomainExample")}
        ]
    )
) as table meta [
    // Connector modal title
    Documentation.Name = Extension.LoadString("ConnectorName")
];

NavTable = (subdomain as text) as table =>
    let
        table = GetJsonData(Url(subdomain, "/api/powerbi")),
        withData = Table.AddColumn(table, "Data", (row) => GetNavItemData(subdomain, row)),
        withItemKind = Table.AddColumn(
            withData, "ItemKind", (row) => if row[IsLeaf] = true then "Table" else "Folder", type text
        ),
        withItemName = Table.AddColumn(
            withItemKind, "ItemName", (row) => if row[IsLeaf] = true then "Table" else "Folder", type text
        ),
        navTable = Table.ToNavigationTable(withItemName, {"Name"}, "Name", "Data", "ItemKind", "ItemName", "IsLeaf")
    in
        navTable;

GetNavItemData = (subdomain as text, navItem as any) =>
    if navItem[IsLeaf] = true then
        GetJsonData(Url(subdomain, navItem[Endpoint]))
    else
        GetNavItemChildren(subdomain, navItem);

GetNavItemChildren = (subdomain as text, navItem as any) =>
    let
        url = Url(subdomain, navItem[Endpoint]),
        table = GetJsonData(url),
        rename = Table.RenameColumns(table, {{navItem[TitleField], "Name"}}),
        withData = Table.AddColumn(rename, "Data", (row) => GetChildData(row, navItem, subdomain)),
        withItemKind = Table.AddColumn(withData, "ItemKind", each "Table", type text),
        withItemName = Table.AddColumn(withItemKind, "ItemName", each "Table", type text),
        withIsLeaf = Table.AddColumn(withItemName, "IsLeaf", each true, type logical),
        navTable = Table.ToNavigationTable(
            withIsLeaf, {navItem[KeyField]}, "Name", "Data", "ItemKind", "ItemName", "IsLeaf"
        )
    in
        navTable;

BuildItemUrl = (item as any, navItem as any, subdomain as text) =>
    let
        withKeyField = Text.Replace(
            navItem[Template], "~KeyField~", Text.From(Record.Field(item, navItem[KeyField]))
        ),
        url = Url(subdomain, withKeyField)
    in
        url;

GetChildData = (item as any, navItem as any, urlBase as text) =>
    let
        url = BuildItemUrl(item, navItem, urlBase),
        maybeEmpty = GetJsonData(url),
        // empty json array has no columns, so we make sure there's always at least one (the keyField)
        // we make sure every child table has this column so that they can be concatenated
        table =
            if Table.HasColumns(maybeEmpty, navItem[KeyField]) then
                maybeEmpty
            else
                Table.AddColumn(
                    maybeEmpty, navItem[KeyField], each Record.Field(item, navItem[KeyField]), type number
                )
    in
        table;

// fetches JSON from the url, expands nested records, and infers the type of columns
// expects the response to be a JSON array
// if the array is empty, the table will be empty (and have no columns)
GetJsonData = (url as text) =>
    let
        response = Web.Contents(url, [
            Headers = [#"User-Agent" = "Power BI Connector " & version]
        ]),
        doc = Json.Document(response),
        table = Table.FromList(doc, Splitter.SplitByNothing()),
        expanded = ExpandRecords(table),
        inferred = SetInferredTypes(expanded)
    in
        inferred;

// the table at this point has a single "Record" column that contains json objects
// it's called recursively to support object properties on that base object
ExpandRecords = (maybeEmpty) =>
    if Table.IsEmpty(maybeEmpty) then
        maybeEmpty
    else
        _ExpandRecordsImpl(maybeEmpty, false, 0);

_ExpandRecordsImpl = (table, autoPrefix as logical, columnIndex as number) =>
    let
        columnName = Table.ColumnNames(table){columnIndex},
        columnContents = Table.Column(table, columnName),
        columnsToExpand = List.Distinct(
            List.Combine(List.Transform(columnContents, each if _ is record then Record.FieldNames(_) else {}))
        ),
        newColumnNames = List.Transform(columnsToExpand, each if autoPrefix then columnName & "." & _ else _),
        canExpandCurrentColumn = List.Count(columnsToExpand) > 0,
        expandedTable =
            if canExpandCurrentColumn then
                Table.ExpandRecordColumn(table, columnName, columnsToExpand, newColumnNames)
            else
                table,
        nextColumnIndex = if canExpandCurrentColumn then columnIndex else columnIndex + 1,
        output =
            if nextColumnIndex > (Table.ColumnCount(expandedTable) - 1) then
                expandedTable
            else
                @_ExpandRecordsImpl(expandedTable, true, nextColumnIndex)
    in
        output;

SetInferredTypes = (maybeEmpty) => if Table.IsEmpty(maybeEmpty) then maybeEmpty else _SetInferredTypesImpl(
    maybeEmpty, 0
);

// Read the contents of each column and try to infer the data type from the non-null values
_SetInferredTypesImpl = (table, columnIndex as number) =>
    let
        columnName = Table.ColumnNames(table){columnIndex},
        columnContents = Table.Column(table, columnName),
        // Only consider the non-null values for the current column
        columnValues = List.RemoveNulls(columnContents),
        // Get the distinct types for the values
        inferredTypes = List.Distinct(
            List.Transform(
                columnContents,
                (value) =>
                    if value is number then
                        type number
                    else if value is text then
                        // If we encounter "XXXX-XX-XX" it's probably a LocalDate field
                        // It would be nice if we could more accurately confirm that Xs are numbers
                        if
                            Text.Length(value) = 10
                            // dirty number check
                            and Text.Upper(value) = Text.Lower(value)
                            // awkward XXXX-XX-XX check (4 chars, dash, 2 chars, dash, 2 chars)
                            and Text.Combine(
                                List.Transform(Text.Split(value, "-"), each Text.From(Text.Length(_))), ""
                            ) = "422"
                        then
                            type date
                        else
                            type text
                    else if value = true or value = false then
                        type logical
                    else
                        type any
            )
        ),
        // If we have only a single inferred type, convert the column to that data type
        updatedTable =
            if List.Count(inferredTypes) = 1 then
                Table.TransformColumnTypes(table, {columnName, inferredTypes{0}})
            else
                table,
        // Process the next column or return the updated table if we're all done
        nextColumnIndex = columnIndex + 1,
        output =
            if nextColumnIndex > (Table.ColumnCount(updatedTable) - 1) then
                updatedTable
            else
                @_SetInferredTypesImpl(updatedTable, nextColumnIndex)
    in
        output;

DevResults.PowerBI = [
    Authentication = [
        OAuth = [
            StartLogin = StartLogin,
            FinishLogin = FinishLogin
        ]
    ],
    TestConnection = (dataSourcePath) =>
        let
            json = Json.Document(dataSourcePath), subdomain = json[subdomain]
        in
            {"DevResults.PowerBI.Contents", subdomain},
    Label = Extension.LoadString("DataSourceLabel")
];

// Navigation conversion function from TripPin tutorial: https://learn.microsoft.com/en-us/power-query/samples/trippin/3-navtables/readme#creating-a-navigation-table
Table.ToNavigationTable = (
    table as table,
    keyColumns as list,
    nameColumn as text,
    dataColumn as text,
    itemKindColumn as text,
    itemNameColumn as text,
    isLeafColumn as text
) as table =>
    let
        tableType = Value.Type(table),
        newTableType = Type.AddTableKey(tableType, keyColumns, true) meta [
            NavigationTable.NameColumn = nameColumn,
            NavigationTable.DataColumn = dataColumn,
            NavigationTable.ItemKindColumn = itemKindColumn,
            Preview.DelayColumn = itemNameColumn,
            NavigationTable.IsLeafColumn = isLeafColumn
        ],
        navigationTable = Value.ReplaceType(table, newTableType)
    in
        navigationTable;

// resourceUrl is JSON of the parameter names and values sent to the Contents function, above
StartLogin = (resourceUrl, state, display) =>
    let
        doc = Json.Document(resourceUrl),
        redirect_uri = RedirectUri(doc[subdomain]),
        AuthorizeUrl = Url(doc[subdomain], "/en/Login?")
            & Uri.BuildQueryString([
                client_id = client_id,
                redirect_uri = redirect_uri
            ])
    in
        [
            LoginUri = AuthorizeUrl,
            CallbackUri = redirect_uri,
            WindowHeight = 750,
            WindowWidth = 800,
            Context = Text.FromBinary(Json.FromValue([subdomain = doc[subdomain]]))
        ];

FinishLogin = (context, callbackUri, state) =>
    let
        Parts = Uri.Parts(callbackUri)[Query]
    in
        TokenMethod(Parts[code], context, state);

TokenMethod = (code, context, state) =>
    let
        doc = Json.Document(context),
        url = Url(doc[subdomain], "/api/oauth2/token"),
        options = [
            Headers = [#"Content-Type" = "application/json"],
            Content = Json.FromValue([code = code, client_id = client_id])
        ],
        response = Web.Contents(url, options),
        Parts = Json.Document(response)
    in
        Parts;

// build a URL based on the subdomain
// note: returns URL with any known "extras" stripped from the domain (e.g. ".mcas.ms" will be removed)
Url = (subdomain as text, path as text) =>
    let
        // 1. make sure we have a full domain (e.g. "leslie" -> "leslie.devresults.com"; "inl.devresults.com.mcas.ms" remains unchanged)
        full_domain = if Text.Contains(subdomain, ".") then subdomain else subdomain & ".devresults.com",
        // 2. we want to use the "bare" domain for all real requests so clean up any known extra stuff from the domain
        // remove MDCA domain wrapper if present (used by INL and possibly other clients in the future)
        domain = Text.Replace(full_domain, ".mcas.ms", ""),
        // 3. validate the path and build the url
        pathWithSlash = if Text.StartsWith(path, "/") then path else "/" & path,
        url = Text.Format("https://#{0}#{1}", {domain, pathWithSlash})
    in
        url;

// build a server expected redirect_uri based on the given subdomain
// we can no longer always use "oauth.devresults.com" because things like MDCA (for INL) can force a rewritten URL like "oauth.devresults.com.mcas.ms"
RedirectUri = (subdomain as text) =>
    let
        // 1. make sure we have a full domain (e.g. "leslie" -> "leslie.devresults.com"; "inl.devresults.com.mcas.ms" remains unchanged)
        domain = if Text.Contains(subdomain, ".") then subdomain else subdomain & ".devresults.com",
        // 2. strip the subdomain, we'll replace it with "oauth"
        without_subdomain = Text.AfterDelimiter(domain, "."),
        // 3. build expected redirect_uri, which should include an "oauth" subdomain and our "/api/oauth/powerbi" route
        // e.g. "https://oauth.devresults.com/api/oauth/powerbi"
        url = Text.Format("https://oauth.#{0}/api/oauth/powerbi", {without_subdomain})
    in
        url;

DevResults.PowerBI.Publish = [
    Category = "Online Services",
    ButtonText = {Extension.LoadString("ButtonTitle"), Extension.LoadString("ButtonHelp")},
    LearnMoreUrl = "https://www.devresults.com/",
    SourceImage = DevResults.PowerBI.Icons,
    SourceTypeImage = DevResults.PowerBI.Icons
];

DevResults.PowerBI.Icons = [
    Icon16 = {
        Extension.Contents("DevResults.PowerBI16.png"),
        Extension.Contents("DevResults.PowerBI20.png"),
        Extension.Contents("DevResults.PowerBI24.png"),
        Extension.Contents("DevResults.PowerBI32.png")
    },
    Icon32 = {
        Extension.Contents("DevResults.PowerBI32.png"),
        Extension.Contents("DevResults.PowerBI40.png"),
        Extension.Contents("DevResults.PowerBI48.png"),
        Extension.Contents("DevResults.PowerBI64.png")
    }
];

